'-------------------------------------------------------------------------------
' (c) Copyright 2000 Rhombus Greenville SC
'Code can be used by others providing this header is included in the source.
'===============================================================================
'Getting started in Real-Time Control using BASCOM                       Part 2
'===============================================================================

'Adding transparent keypad input

'To maintain consistent processing of the I/O tasks in Part 1 we cannot leave
' the main loop, start scanning a keypad and wait for user input. Any keypad
' activity needs to be flagged and brought into the main loop by one means or
' another - and at a pre-determined priority level with respect to other tasks.

'A key encoder chip could take care of full-time scanning, separation of
' multiple keys, debouncing, and we would just watch for its data ready flag.
' For example a 74C922 would need 5 port pins from an MCU or a decode of one
' address for an MPU, to handle 16 keys - they are somewhat more expensive than
' 51 micro itself, but would serve our purpose well.

'An alternative is shown below where the existing Timer0 ISR takes on the
' additional task of sinking just one of the 4 row signals connected to a
' keypad matrix. Everytime the Scan-tics reach a specific count (shown for 10x2
' =20mS) then the next row becomes active low, until the columns indicate a key
' is depressed. Then the active row is maintained and the columns must read the
' same after a further 20mS in order for the key to be accepted. That takes 3
' extra port pins for a total of 8, but no additional logic is required. It
' also scans full-time.

'Either solution will allow detection of an active key in the main loop. The
' code shown for key detection below becomes 'If Key_flag Then Goto Key_10' when
' using an MCU with a 74C922. The Row_col was read and saved at the 20mS point,
' the 922 also saved after debounce and the operator will not be sensitive to a
' few hundred mS of other tasks being completed prior to acting on the last key.
' So key detection receives one of the lowest priorities and will be close to
' the bottom of the loop.

'If fixed and pre-determined functions were associated with each key, then
' keypad input would already be complete. Each key would have an associated
' Goto and they would all return to the top of the main loop (standard practice,
' and no stack used). They may only require one or two bits to be set/reset and
' occupy just a few uSecs. But often that one key will be the first step to an
' operator dialog taking seconds or even minutes and each operator response
' may well consist of several keys as a numeric value is entered and edited.

'That operator dialog routine needs to be a subroutine to allow for multiple
' callers, yet it must immediately hand execution back to the main loop to
' satisfy the needs of other tasks, and further to receive its own key input.
' The first such key pressed will exit the main loop in a manner identical to
' the function key that started the dialog. However, its non-function code will
' allow it be handled differently. When complete, again the key code common to
' both function and dialog keys will be able to distinguish between the cases
' and this time will execute a Return rather than Goto top of main loop. Then
' the calling routine itself will pass the execution to the top of the main loop
' once it has retrieved its dialog data.

' The common code could be included in the main loop, but it will become lengthy
' if code for key masks,lengths,min/max values etc are added, and that would
' detract from the loop's clarity.

'At this stage we can now enter new data for each of the timers in parallel
' with normal operation of the 'machine'. Using a scope on P3.5 as a reference.
' the only visual influence of keyboard activity will be when the final Enter
' key causes the trace to snap to the new timing.

'Part 3 will add transparent host communication.

'----- INITIALISATION

Dim Timers(4) As Byte , Isr_temp As Byte , Temp1 As Byte , Temp2 As Byte
Dim Timer_tics As Byte , Key_tics As Byte , Io_tics As Byte
Dim Old_rowcol As Byte , Row_col As Byte , Key_val As Byte
Dim Key_len As Byte , Key$ As String * 5
Dim New_key As Bit , No_key As Bit , Pad_active As Bit , Io_flag As Bit
Dim Set_pt1 As Byte , Set_pt2 As Byte , Set_pt3 As Byte , Set_pt4 As Byte

Set_pt1 = 20 : Set_pt2 = 15 : Set_pt3 = 10 : Set_pt4 = 5
Row_col = &HFF
Key_len = 3
Const 2ms = -1793                                   ' (2/1000)*(11059200/12)-50reload
Declare Sub Key_rtn()
$baud = 9600

Config Timer0 = Timer , Gate = Internal , Mode = 1       '16 bit, code reloads
Counter0 = 2ms
On Timer0 Timer_0_int
Enable Interrupts                                   'enable the use of interrupts
Enable Timer0

Priority Set Timer0                                 'highest priority
Start Timer0

'For demo purposes the 'machine' consists of 4 timers cycling & driving 4 port
' pins. There are 4 keypad function keys that initiate a dialog sequence with the
' machine operator to setup new timer values in real-time whilst the 'machine'
' continues normally.
'The keypad is scanned continously in the background and collects the user's
' input without interrupting the control of the machine - a scope can be used
' to monitor the machine's cycling pins as the timers are changed.

'----- MAIN LOOP (task priorities in descending order)

Main_loop:
Do
'--Test for I/O prompt
   If Io_flag = 1 Then Goto Io_control              'Dummy machine control

'--Test for waiting Host messages                   'Will be added in Part 3
   'If Msg_flag Then Goto Msg_rtn

' (ADDED with Part2)
'--Test for keypad activity
   If New_key = 0 Then                              'Look for new keys
      Temp1 = Row_col And &H0F                      '4 LSbits represent 4 Colmns
      If Temp1 <> &H0F Then                         'If a key is 'Active'
         If Old_rowcol <> Row_col Then              ' and a new pattern
            Old_rowcol = Row_col                    ' then record first stage &
            Reset No_key
            Row_col = &HFF                          ' ensure Row_col not re-used
         Else
             Goto Key_10                            'Now 2nd stage & definitely
         End If
      End If                                        ' have a key to process
   End If
Loop

 ' (ADDED with Part2)
'----- KEYBOARD DECODER

Sub Key_rtn                                         'Call from any dialog routine
   Key$ = ""
   'LCD ....
   Pad_active = 1
   Goto Main_loop                                   'Get next key

'-- From here upto Key_return could be included as part of main loop - but see text
    'Each new character causes a 'Jump' from main loop to Key_10
Key_10:
   Temp1 = Lookup(temp1 , Xlate_col)                'Translate Column to = 0,1,2,3
   Temp2 = Row_col And &HF0                         ' isolate Row to = 70,B0,D0,EO
   Select Case Temp2
      Case &H70 : Key_val = Lookup(temp1 , Row0)    ' then use that Row's lookup
      Case &HB0 : Key_val = Lookup(temp1 , Row1)
      Case &HD0 : Key_val = Lookup(temp1 , Row2)
      Case &HE0 : Key_val = Lookup(temp1 , Row3)
      Case Else :
   End Select
   Set New_key                                      ' and avoid repeat keys
   'Function Key cases
   If Temp1 = 0 Then                                'Left hand column of 4 keys
      If Pad_active = 0 Then                        'Reject new Funxs during Dialog
         On Key_val Goto Edit_sp1 , Edit_sp2 , Edit_sp3 , Edit_sp4
      Else
         Goto Main_loop
      End If
   Else
   'Dialog key cases
   'Filtering shown for backspace & CR only - also Max_len
      Temp1 = Len(key$)
      If Key_val <> 13 Then                         'CR terminates dialog in example
         If Key_val = 08 Then                       ' and falls thru to Key_return
            Temp1 = Temp1 - 1
            If Temp1 < 0 Then Temp1 = 0
            Key$ = Left(key$ , Temp1)
         Elseif Temp1 < Key_len Then
            Key$ = Key$ + Chr(key_val)
         End If
         Print Key$;
         'Locate Y , X : Lcd Key$
         Goto Main_loop
      Elseif Temp1 = 0 Then
         Goto Main_loop
      End If
   End If
   'Goto Key_return
'-- End of code that could be included in the main loop - would need immediately
'    above line additionally

Key_return:
   Print "CR"
   Pad_active = 0
   Return
End Sub

' (ADDED with Part2)
'----- FUNCTIONS USING DIALOG (stripped down to keys only - (& no LCD to test with))

Edit_sp1:
   'LCD prompt for operator                         'Enter new Setpoint Routine
   Print "SP1"
   Key_rtn                                          'Returns with an Enter Key
   Set_pt1 = Val(key$)                              'Makes new value effective
                                                    ' when timer is loaded next
   'Wasn't Called so can 'Jump' to highest priority task
   Goto Main_loop:

Edit_sp2:                                           ' As above
   'LCD prompt for operator
   Print "SP2"
   Key_rtn
   Set_pt2 = Val(key$)
   Goto Main_loop:

Edit_sp3:                                           ' As above
   'LCD prompt for operator
   Print "SP3"
   Key_rtn
   Set_pt3 = Val(key$)
   Goto Main_loop:

Edit_sp4:                                           ' As above
   'LCD prompt for operator
   Print "SP4"
   Key_rtn
   Set_pt4 = Val(key$)
   Goto Main_loop:

  '----- MACHINE CONTROL
Io_control:
   'The 3 stages of Read_Inputs/Process/Write_Outputs
   'All inputs here are Timer values and hence readily available internally.
   'The processing here is merely to watch for Timers(1) to count down to zero
   ' and then reset its associated P3.5. On the next pass (when Io_flag is found
   ' set in the Main_loop) we reload all 4 timers with their individual values &
   ' set their individual flags.
   'All subsequent passes with Timer(1)<>0 has pins P3.2.3.4 updated to reflect
   ' when their counts have reached zero.
   'That is not a separate Output stage as such, but if we had remote I/O then
   ' the equivalents of P3.2.. would be internal bits and once all processing
   ' was complete, we would output them as a distinct & separate operation as in
   ' a PLC.
   Io_flag = 0
   If Timers(1) = 0 Then
      If P3.5 = 1 Then                              'Dont restart till 1 pass later
         Reset P3.5
      Else
         Timers(1) = Set_pt1 : Set P3.5             'In Part1 we had fixed values
         Timers(2) = Set_pt2 : Set P3.2             ' now we need vars Set_ptx to
         Timers(3) = Set_pt3 : Set P3.3             ' allow the KeyPad input.
         Timers(4) = Set_pt4 : Set P3.4
      End If
   Else
      If Timers(2) = 0 Then Reset P3.2              'Could avoid multiple resets
      If Timers(3) = 0 Then Reset P3.3              ' but no real savings
      If Timers(4) = 0 Then Reset P3.4
   End If
   Goto Main_loop

   'We return at the highest priority level ignoring the tasks below this one
   ' but know that they will be reached on the next pass, now that the Io_flag
   ' is reset. *Except* if the interval set between scans is not realistic: eg
   ' if the Control routines take 50mS and the Io_flag is set every 40mS then
   ' all other functions will be blocked out. Timing the different functions in
   ' the Main_loop and allocating realistic intervals is not a difficult task.


'----- INTERRUPT DRIVEN TIMERS (Keyboard Scanning added to code of Part 1)

Timer_0_int:
   Counter0 = 2ms
   Start Timer0
   Set Io_flag
   Inc Timer_tics
   If Timer_tics => 1 Then                          '01 and fast to allow scoping
      Timer_tics = 0                                '05/50 = 10mS/100mS more typical
      For Isr_temp = 1 To 4
         If Timers(isr_temp) <> 0 Then Decr Timers(isr_temp)
      Next

   ' (ADDED with Part2)
   'Keyboard Scanning
   End If
   Inc Key_tics
   If Key_tics >= 25 Then
      Key_tics = 0
      Row_col = P1
      Isr_temp = Row_col And &H0F                   '&H0F says no Cols driven low
      If Isr_temp = &H0F Then                       'Continue scanning whilst no key
         If No_key = 1 Then Reset New_key Else Set No_key
         Select Case Row_col                        'Rows rotate '0' Cols always I/Ps
            Case &H7F : Row_col = &HBF              '0111 1111 Row0->1 1011 1111
            Case &HBF : Row_col = &HDF              '1011 1111 Row1->2 1101 1111
            Case &HDF : Row_col = &HEF              '1101 1111 Row2->3 1110 1111
            Case &HEF : Row_col = &H7F              '1110 1111 Row3->0 0111 1111
            Case Else : Row_col = &H7F
         End Select
         P1 = Row_col
       End If
   End If
Return

Xlate_col:
Data 0 , 0 , 0 , 0 , 0 , 0 , 0 , 0 , 1 , 1 , 1 , 1 , 2 , 2 , 3
'Col# *  *   *   *   *   *   *   0   *   *   *   1   *   2   3  Posn15not possible
' '*' due multiple keys
Row0:
Data 00 , 49 , 50 , 51                              'F0  1  2  3
Row1:
Data 01 , 52 , 53 , 54                              'F1  4  5  6
Row2:
Data 02 , 55 , 56 , 57                              'F2  7  8  9
Row3:
Data 03 , 08 , 48 , 13                              'F3  BS 0 CR

'Above lookup tables assume following relationship between port bits and keyboard
'Rows/Columns:
'Port Bits:            7 6 5 4 3 2 1 0
'Rows Top to Bottom:   0 1 2 3
'Columns Left to Right         0 1 2(3)
'Code assumes Port.0 bit to be high when 4x3 pad

'Xlate_col shown for leftmost column taking priority when multiple keys on same
' row are depressed.
'Multiple keys on same column are resolved by the scanning - it halts on first
' key scanned, going down.
'Good for 3 or 4 column key pads - extend Xlate_col with 7'0s'to avoid bad data
' with multiple key cases